from django import forms
from pytest import raises

import graphene
from graphene_django import DjangoObjectType

from ...tests.models import CHOICES, Film, Reporter
from ..types import DjangoFormInputObjectType

# Reporter a_choice CHOICES = ((1, "this"), (2, _("that")))
THIS = CHOICES[0][0]
THIS_ON_CLIENT_CONVERTED = "A_1"

# Film genre choices=[("do", "Documentary"), ("ac", "Action"), ("ot", "Other")],
DOCUMENTARY = "do"
DOCUMENTARY_ON_CLIENT_CONVERTED = "DO"


class FilmForm(forms.ModelForm):
    class Meta:
        model = Film
        exclude = ()


class ReporterType(DjangoObjectType):
    class Meta:
        model = Reporter
        fields = "__all__"


class ReporterForm(forms.ModelForm):
    class Meta:
        model = Reporter
        exclude = ("pets", "email", "fans")


class MyForm(forms.Form):
    text_field = forms.CharField()
    int_field = forms.IntegerField()


def test_needs_form_class():
    with raises(Exception) as exc:

        class MyInputType(DjangoFormInputObjectType):
            pass

    assert exc.value.args[0] == "form_class is required for DjangoFormInputObjectType"


def test_type_from_modelform_has_input_fields():
    class ReporterInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = ReporterForm
            only_fields = ("first_name", "last_name", "a_choice")

    fields = ["first_name", "last_name", "a_choice", "id"]
    assert all(f in ReporterInputType._meta.fields for f in fields)


def test_type_from_form_has_input_fields():
    class MyFormInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = MyForm

    fields = ["text_field", "int_field", "id"]
    assert all(f in MyFormInputType._meta.fields for f in fields)


def test_type_custom_id_field():
    class MyFormInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = MyForm
            add_id_field_name = "pk"

    fields = ["text_field", "int_field", "pk"]
    assert all(f in MyFormInputType._meta.fields for f in fields)
    assert MyFormInputType._meta.fields["pk"].type is graphene.ID


def test_type_custom_id_field_type():
    class MyFormInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = MyForm
            add_id_field_name = "pk"
            add_id_field_type = graphene.String(required=False)

    fields = ["text_field", "int_field", "pk"]
    assert all(f in MyFormInputType._meta.fields for f in fields)
    assert MyFormInputType._meta.fields["pk"].type is graphene.String


class MockQuery(graphene.ObjectType):
    a = graphene.String()


def test_mutation_with_form_djangoforminputtype():
    class MyFormInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = MyForm

    class MyFormMutation(graphene.Mutation):
        class Arguments:
            form_data = MyFormInputType(required=True)

        result = graphene.Boolean()

        def mutate(_root, _info, form_data):
            form = MyForm(data=form_data)
            if form.is_valid():
                result = form.cleaned_data == {
                    "text_field": "text",
                    "int_field": 777,
                }
                return MyFormMutation(result=result)
            return MyFormMutation(result=False)

    class Mutation(graphene.ObjectType):
        myForm_mutation = MyFormMutation.Field()

    schema = graphene.Schema(query=MockQuery, mutation=Mutation)

    result = schema.execute(
        """ mutation MyFormMutation($formData: MyFormInputType!) {
            myFormMutation(formData: $formData) {
                result
            }
        }
        """,
        variable_values={"formData": {"textField": "text", "intField": 777}},
    )
    assert result.errors is None
    assert result.data == {"myFormMutation": {"result": True}}


def test_mutation_with_modelform_djangoforminputtype():
    class ReporterInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = ReporterForm
            object_type = ReporterType
            only_fields = ("first_name", "last_name", "a_choice")

    class ReporterMutation(graphene.Mutation):
        class Arguments:
            reporter_data = ReporterInputType(required=True)

        result = graphene.Field(ReporterType)

        def mutate(_root, _info, reporter_data):
            reporter = Reporter.objects.get(pk=reporter_data.id)
            form = ReporterForm(data=reporter_data, instance=reporter)
            if form.is_valid():
                reporter = form.save()
                return ReporterMutation(result=reporter)

            return ReporterMutation(result=None)

    class Mutation(graphene.ObjectType):
        report_mutation = ReporterMutation.Field()

    schema = graphene.Schema(query=MockQuery, mutation=Mutation)

    reporter = Reporter.objects.create(
        first_name="Bob", last_name="Roberts", a_choice=THIS
    )

    result = schema.execute(
        """ mutation ReportMutation($reporterData: ReporterInputType!) {
            reportMutation(reporterData: $reporterData) {
                result {
                    id,
                    firstName,
                    lastName,
                    aChoice
                }
            }
        }
        """,
        variable_values={
            "reporterData": {
                "id": reporter.pk,
                "firstName": "Dave",
                "lastName": "Smith",
                "aChoice": THIS_ON_CLIENT_CONVERTED,
            }
        },
    )
    assert result.errors is None
    assert result.data["reportMutation"]["result"] == {
        "id": "1",
        "firstName": "Dave",
        "lastName": "Smith",
        "aChoice": THIS_ON_CLIENT_CONVERTED,
    }
    assert Reporter.objects.count() == 1
    reporter.refresh_from_db()
    assert reporter.first_name == "Dave"


def reporter_enum_convert_mutation_result(
    ReporterInputType, choice_val_on_client=THIS_ON_CLIENT_CONVERTED
):
    class ReporterMutation(graphene.Mutation):
        class Arguments:
            reporter = ReporterInputType(required=True)

        result_str = graphene.String()
        result_int = graphene.Int()

        def mutate(_root, _info, reporter):
            if isinstance(reporter.a_choice, int) or reporter.a_choice.isdigit():
                return ReporterMutation(result_int=reporter.a_choice, result_str=None)
            return ReporterMutation(result_int=None, result_str=reporter.a_choice)

    class Mutation(graphene.ObjectType):
        report_mutation = ReporterMutation.Field()

    schema = graphene.Schema(query=MockQuery, mutation=Mutation)

    return schema.execute(
        """ mutation ReportMutation($reporter: ReporterInputType!) {
            reportMutation(reporter: $reporter) {
                resultStr,
                resultInt
            }
        }
        """,
        variable_values={"reporter": {"aChoice": choice_val_on_client}},
    )


def test_enum_not_converted():
    class ReporterInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = ReporterForm
            only_fields = ("a_choice",)

    result = reporter_enum_convert_mutation_result(ReporterInputType)
    assert result.errors is None
    assert result.data["reportMutation"]["resultStr"] == THIS_ON_CLIENT_CONVERTED
    assert result.data["reportMutation"]["resultInt"] is None
    assert ReporterInputType._meta.fields["a_choice"].type is graphene.String


def test_enum_is_converted_to_original():
    class ReporterInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = ReporterForm
            object_type = ReporterType
            only_fields = ("a_choice",)

    result = reporter_enum_convert_mutation_result(ReporterInputType)
    assert result.errors is None
    assert result.data["reportMutation"]["resultInt"] == THIS
    assert result.data["reportMutation"]["resultStr"] is None
    assert (
        ReporterInputType._meta.fields["a_choice"].type.__name__
        == "AChoiceEnumBackConvString"
    )


def test_convert_choices_to_enum_is_false_and_field_type_as_in_model():
    class ReporterTypeNotConvertChoices(DjangoObjectType):
        class Meta:
            model = Reporter
            convert_choices_to_enum = False
            fields = "__all__"

    class ReporterInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = ReporterForm
            object_type = ReporterTypeNotConvertChoices
            only_fields = ("a_choice",)

    result = reporter_enum_convert_mutation_result(ReporterInputType, THIS)
    assert result.errors is None
    assert result.data["reportMutation"]["resultInt"] == THIS
    assert result.data["reportMutation"]["resultStr"] is None
    assert ReporterInputType._meta.fields["a_choice"].type is graphene.Int


def enum_convert_mutation_result_film(FilmInputType):
    class FilmMutation(graphene.Mutation):
        class Arguments:
            film = FilmInputType(required=True)

        result = graphene.String()

        def mutate(_root, _info, film):
            return FilmMutation(result=film.genre)

    class Mutation(graphene.ObjectType):
        film_mutation = FilmMutation.Field()

    schema = graphene.Schema(query=MockQuery, mutation=Mutation)

    return schema.execute(
        """ mutation FilmMutation($film: FilmInputType!) {
            filmMutation(film: $film) {
                result
            }
        }
        """,
        variable_values={"film": {"genre": DOCUMENTARY_ON_CLIENT_CONVERTED}},
    )


def test_enum_not_converted_required_non_number():
    class FilmInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = FilmForm
            only_fields = ("genre",)

    result = enum_convert_mutation_result_film(FilmInputType)
    assert result.errors is None
    assert result.data["filmMutation"]["result"] == DOCUMENTARY_ON_CLIENT_CONVERTED


def test_enum_is_converted_to_original_required_non_number():
    class FilmType(DjangoObjectType):
        class Meta:
            model = Film
            fields = "__all__"

    class FilmInputType(DjangoFormInputObjectType):
        class Meta:
            form_class = FilmForm
            object_type = FilmType
            only_fields = ("genre",)

    result = enum_convert_mutation_result_film(FilmInputType)
    assert result.errors is None
    assert result.data["filmMutation"]["result"] == DOCUMENTARY



class Registry:
    def __init__(self):
        self._registry = {}
        self._field_registry = {}

    def register(self, cls):
        from .types import DjangoObjectType

        assert issubclass(
            cls, DjangoObjectType
        ), f'Only DjangoObjectTypes can be registered, received "{cls.__name__}"'
        assert cls._meta.registry == self, "Registry for a Model have to match."
        # assert self.get_type_for_model(cls._meta.model) == cls, (
        #     'Multiple DjangoObjectTypes registered for "{}"'.format(cls._meta.model)
        # )
        if not getattr(cls._meta, "skip_registry", False):
            self._registry[cls._meta.model] = cls

    def get_type_for_model(self, model):
        return self._registry.get(model)

    def register_converted_field(self, field, converted):
        self._field_registry[field] = converted

    def get_converted_field(self, field):
        return self._field_registry.get(field)


registry = None


def get_global_registry():
    global registry
    if not registry:
        registry = Registry()
    return registry


def reset_global_registry():
    global registry
    registry = None



import datetime
import re

import pytest
from django.db.models import Count, Model, Prefetch

from graphene import List, NonNull, ObjectType, Schema, String
from graphene.relay import Node

from ..fields import DjangoConnectionField, DjangoListField
from ..types import DjangoObjectType
from .models import (
    Article as ArticleModel,
    Film as FilmModel,
    FilmDetails as FilmDetailsModel,
    Person as PersonModel,
    Reporter as ReporterModel,
)


class TestDjangoListField:
    def test_only_django_object_types(self):
        class Query(ObjectType):
            something = DjangoListField(String)

        with pytest.raises(TypeError) as excinfo:
            Schema(query=Query)

        assert (
            "Query fields cannot be resolved. DjangoListField only accepts DjangoObjectType types as underlying type"
            in str(excinfo.value)
        )

    def test_only_import_paths(self):
        list_field = DjangoListField("graphene_django.tests.schema.Human")
        from .schema import Human

        assert list_field._type.of_type.of_type is Human

    def test_non_null_type(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name",)

        list_field = DjangoListField(NonNull(Reporter))

        assert isinstance(list_field.type, List)
        assert isinstance(list_field.type.of_type, NonNull)
        assert list_field.type.of_type.of_type is Reporter

    def test_get_django_model(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name",)

        list_field = DjangoListField(Reporter)
        assert list_field.model is ReporterModel

    def test_list_field_default_queryset(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name",)

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "reporters": [{"firstName": "Tara"}, {"firstName": "Debra"}]
        }

    def test_list_field_queryset_is_not_cached(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name",)

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        result = schema.execute(query)
        assert not result.errors
        assert result.data == {"reporters": []}

        ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "reporters": [{"firstName": "Tara"}, {"firstName": "Debra"}]
        }

    def test_override_resolver(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name",)

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

            def resolve_reporters(_, info):
                return ReporterModel.objects.filter(first_name="Tara")

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {"reporters": [{"firstName": "Tara"}]}

    def test_nested_list_field(self):
        class Article(DjangoObjectType):
            class Meta:
                model = ArticleModel
                fields = ("headline",)

        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                    articles {
                        headline
                    }
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )
        ArticleModel.objects.create(
            headline="Not so good news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "reporters": [
                {
                    "firstName": "Tara",
                    "articles": [
                        {"headline": "Amazing news"},
                        {"headline": "Not so good news"},
                    ],
                },
                {"firstName": "Debra", "articles": []},
            ]
        }

    def test_override_resolver_nested_list_field(self):
        class Article(DjangoObjectType):
            class Meta:
                model = ArticleModel
                fields = ("headline",)

        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

            def resolve_articles(reporter, info):
                return reporter.articles.filter(headline__contains="Amazing")

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                    articles {
                        headline
                    }
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )
        ArticleModel.objects.create(
            headline="Not so good news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "reporters": [
                {"firstName": "Tara", "articles": [{"headline": "Amazing news"}]},
                {"firstName": "Debra", "articles": []},
            ]
        }

    def test_same_type_nested_list_field(self):
        class Person(DjangoObjectType):
            class Meta:
                model = PersonModel
                fields = ("name", "parent")

            children = DjangoListField(lambda: Person)

        class Query(ObjectType):
            persons = DjangoListField(Person)

        schema = Schema(query=Query)

        query = """
            query {
                persons {
                    name
                    children {
                        name
                    }
                }
            }
        """

        p1 = PersonModel.objects.create(name="Tara")
        PersonModel.objects.create(name="Debra")

        PersonModel.objects.create(
            name="Toto",
            parent=p1,
        )
        PersonModel.objects.create(
            name="Tata",
            parent=p1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "persons": [
                {
                    "name": "Tara",
                    "children": [
                        {"name": "Toto"},
                        {"name": "Tata"},
                    ],
                },
                {
                    "name": "Debra",
                    "children": [],
                },
                {
                    "name": "Toto",
                    "children": [],
                },
                {
                    "name": "Tata",
                    "children": [],
                },
            ]
        }

    def test_get_queryset_filter(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

            @classmethod
            def get_queryset(cls, queryset, info):
                # Only get reporters with at least 1 article
                return queryset.annotate(article_count=Count("articles")).filter(
                    article_count__gt=0
                )

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

            def resolve_reporters(_, info):
                return ReporterModel.objects.all()

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {"reporters": [{"firstName": "Tara"}]}

    def test_resolve_list(self):
        """Resolving a plain list should work (and not call get_queryset)"""

        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

            @classmethod
            def get_queryset(cls, queryset, info):
                # Only get reporters with at least 1 article
                return queryset.annotate(article_count=Count("articles")).filter(
                    article_count__gt=0
                )

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

            def resolve_reporters(_, info):
                return [ReporterModel.objects.get(first_name="Debra")]

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {"reporters": [{"firstName": "Debra"}]}

    def test_get_queryset_foreign_key(self):
        class Article(DjangoObjectType):
            class Meta:
                model = ArticleModel
                fields = ("headline",)

            @classmethod
            def get_queryset(cls, queryset, info):
                # Rose tinted glasses
                return queryset.exclude(headline__contains="Not so good")

        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

        class Query(ObjectType):
            reporters = DjangoListField(Reporter)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                    articles {
                        headline
                    }
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )
        ArticleModel.objects.create(
            headline="Not so good news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "reporters": [
                {"firstName": "Tara", "articles": [{"headline": "Amazing news"}]},
                {"firstName": "Debra", "articles": []},
            ]
        }

    def test_resolve_list_external_resolver(self):
        """Resolving a plain list from external resolver should work (and not call get_queryset)"""

        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

            @classmethod
            def get_queryset(cls, queryset, info):
                # Only get reporters with at least 1 article
                return queryset.annotate(article_count=Count("articles")).filter(
                    article_count__gt=0
                )

        def resolve_reporters(_, info):
            return [ReporterModel.objects.get(first_name="Debra")]

        class Query(ObjectType):
            reporters = DjangoListField(Reporter, resolver=resolve_reporters)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {"reporters": [{"firstName": "Debra"}]}

    def test_get_queryset_filter_external_resolver(self):
        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles")

            @classmethod
            def get_queryset(cls, queryset, info):
                # Only get reporters with at least 1 article
                return queryset.annotate(article_count=Count("articles")).filter(
                    article_count__gt=0
                )

        def resolve_reporters(_, info):
            return ReporterModel.objects.all()

        class Query(ObjectType):
            reporters = DjangoListField(Reporter, resolver=resolve_reporters)

        schema = Schema(query=Query)

        query = """
            query {
                reporters {
                    firstName
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        result = schema.execute(query)

        assert not result.errors
        assert result.data == {"reporters": [{"firstName": "Tara"}]}

    def test_select_related_and_prefetch_related_are_respected(
        self, django_assert_num_queries
    ):
        class Article(DjangoObjectType):
            class Meta:
                model = ArticleModel
                fields = ("headline", "editor", "reporter")

        class Film(DjangoObjectType):
            class Meta:
                model = FilmModel
                fields = ("genre", "details")

        class FilmDetail(DjangoObjectType):
            class Meta:
                model = FilmDetailsModel
                fields = ("location",)

        class Reporter(DjangoObjectType):
            class Meta:
                model = ReporterModel
                fields = ("first_name", "articles", "films")

        class Query(ObjectType):
            articles = DjangoListField(Article)

            @staticmethod
            def resolve_articles(root, info):
                # Optimize for querying associated editors and reporters, and the films and film
                # details of those reporters. This is similar to what would happen using a library
                # like https://github.com/tfoxy/graphene-django-optimizer for a query like the one
                # below (albeit simplified and hardcoded here).
                return ArticleModel.objects.select_related(
                    "editor", "reporter"
                ).prefetch_related(
                    Prefetch(
                        "reporter__films",
                        queryset=FilmModel.objects.select_related("details"),
                    ),
                )

        schema = Schema(query=Query)

        query = """
            query {
                articles {
                    headline

                    editor {
                        firstName
                    }

                    reporter {
                        firstName

                        films {
                            genre

                            details {
                                location
                            }
                        }
                    }
                }
            }
        """

        r1 = ReporterModel.objects.create(first_name="Tara", last_name="West")
        r2 = ReporterModel.objects.create(first_name="Debra", last_name="Payne")

        ArticleModel.objects.create(
            headline="Amazing news",
            reporter=r1,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r2,
        )
        ArticleModel.objects.create(
            headline="Not so good news",
            reporter=r2,
            pub_date=datetime.date.today(),
            pub_date_time=datetime.datetime.now(),
            editor=r1,
        )

        film1 = FilmModel.objects.create(genre="ac")
        film2 = FilmModel.objects.create(genre="ot")
        film3 = FilmModel.objects.create(genre="do")
        FilmDetailsModel.objects.create(location="Hollywood", film=film1)
        FilmDetailsModel.objects.create(location="Antarctica", film=film3)
        r1.films.add(film1, film2)
        r2.films.add(film3)

        # We expect 2 queries to be performed based on the above resolver definition: one for all
        # articles joined with the reporters model (for associated editors and reporters), and one
        # for the films prefetch (which includes its `select_related` JOIN logic in its queryset)
        with django_assert_num_queries(2) as captured:
            result = schema.execute(query)

        assert not result.errors
        assert result.data == {
            "articles": [
                {
                    "headline": "Amazing news",
                    "editor": {"firstName": "Debra"},
                    "reporter": {
                        "firstName": "Tara",
                        "films": [
                            {"genre": "AC", "details": {"location": "Hollywood"}},
                            {"genre": "OT", "details": None},
                        ],
                    },
                },
                {
                    "headline": "Not so good news",
                    "editor": {"firstName": "Tara"},
                    "reporter": {
                        "firstName": "Debra",
                        "films": [
                            {"genre": "DO", "details": {"location": "Antarctica"}},
                        ],
                    },
                },
            ]
        }

        assert len(captured.captured_queries) == 2  # Sanity-check

        # First we should have queried for all articles in a single query, joining on the reporters
        # model (for the editors and reporters ForeignKeys)
        assert re.match(
            r'SELECT .* "tests_article" INNER JOIN "tests_reporter"',
            captured.captured_queries[0]["sql"],
        )

        # Then we should have queried for all of the films of all reporters, joined with the film
        # details for each film, using a single query
        assert re.match(
            r'SELECT .* FROM "tests_film" INNER JOIN "tests_film_reporters" .* LEFT OUTER JOIN "tests_filmdetails"',
            captured.captured_queries[1]["sql"],
        )


class TestDjangoConnectionField:
    def test_model_ordering_assertion(self):
        class Chaos(Model):
            class Meta:
                app_label = "test"

        class ChaosType(DjangoObjectType):
            class Meta:
                model = Chaos
                interfaces = (Node,)

        class Query(ObjectType):
            chaos = DjangoConnectionField(ChaosType)

        with pytest.raises(
            TypeError,
            match=r"Django model test\.Chaos has to have a default ordering to be used in a Connection\.",
        ):
            Schema(query=Query)

    def test_only_django_object_types(self):
        class Query(ObjectType):
            something = DjangoConnectionField(String)

        with pytest.raises(
            TypeError,
            match="DjangoConnectionField only accepts DjangoObjectType types as underlying type",
        ):
            Schema(query=Query)



import pytest

from graphene_django.settings import graphene_settings as gsettings

from .registry import reset_global_registry


@pytest.fixture(autouse=True)
def reset_registry_fixture(db):
    yield None
    reset_global_registry()


@pytest.fixture()
def graphene_settings():
    settings = dict(gsettings.__dict__)
    yield gsettings
    gsettings.__dict__ = settings



from django.db import models
from django.utils.translation import gettext_lazy as _

CHOICES = ((1, "this"), (2, _("that")))


class Person(models.Model):
    class Meta:
        ordering = ["pk"]

    name = models.CharField(max_length=30)
    parent = models.ForeignKey(
        "self", on_delete=models.CASCADE, null=True, blank=True, related_name="children"
    )


class Pet(models.Model):
    class Meta:
        ordering = ["pk"]

    name = models.CharField(max_length=30)
    age = models.PositiveIntegerField()
    owner = models.ForeignKey(
        "Person", on_delete=models.CASCADE, null=True, blank=True, related_name="pets"
    )


class FilmDetails(models.Model):
    location = models.CharField(max_length=30)
    film = models.OneToOneField(
        "Film",
        on_delete=models.CASCADE,
        related_name="details",
        null=True,
        blank=True,
    )


class Film(models.Model):
    class Meta:
        ordering = ["pk"]

    genre = models.CharField(
        max_length=2,
        help_text="Genre",
        choices=[("do", "Documentary"), ("ac", "Action"), ("ot", "Other")],
        default="ot",
    )
    reporters = models.ManyToManyField("Reporter", related_name="films")


class DoeReporterManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(last_name="Doe")


class Reporter(models.Model):
    class Meta:
        ordering = ["pk"]

    first_name = models.CharField(max_length=30)
    last_name = models.CharField(max_length=30)
    email = models.EmailField()
    pets = models.ManyToManyField("self")
    a_choice = models.IntegerField(choices=CHOICES, null=True, blank=True)
    objects = models.Manager()
    doe_objects = DoeReporterManager()
    fans = models.ManyToManyField(Person)

    reporter_type = models.IntegerField(
        "Reporter Type",
        null=True,
        blank=True,
        choices=[(1, "Regular"), (2, "CNN Reporter")],
    )

    def __str__(self):  # __unicode__ on Python 2
        return f"{self.first_name} {self.last_name}"

    def __init__(self, *args, **kwargs):
        """
        Override the init method so that during runtime, Django
        can know that this object can be a CNNReporter by casting
        it to the proxy model. Otherwise, as far as Django knows,
        when a CNNReporter is pulled from the database, it is still
        of type Reporter. This was added to test proxy model support.
        """
        super().__init__(*args, **kwargs)
        if self.reporter_type == 2:  # quick and dirty way without enums
            self.__class__ = CNNReporter

    def some_method(self):
        return 123


class CNNReporterManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(reporter_type=2)


class CNNReporter(Reporter):
    """
    This class is a proxy model for Reporter, used for testing
    proxy model support
    """

    class Meta:
        proxy = True

    objects = CNNReporterManager()


class APNewsReporter(Reporter):
    """
    This class only inherits from Reporter for testing multi table inheritance
    similar to what you'd see in django-polymorphic
    """

    alias = models.CharField(max_length=30)
    objects = models.Manager()


class Article(models.Model):
    headline = models.CharField(max_length=100)
    pub_date = models.DateField(auto_now_add=True)
    pub_date_time = models.DateTimeField(auto_now_add=True)
    reporter = models.ForeignKey(
        Reporter, on_delete=models.CASCADE, related_name="articles"
    )
    editor = models.ForeignKey(
        Reporter, on_delete=models.CASCADE, related_name="edited_articles_+"
    )
    lang = models.CharField(
        max_length=2,
        help_text="Language",
        choices=[("es", "Spanish"), ("en", "English")],
        default="es",
    )
    importance = models.IntegerField(
        "Importance",
        null=True,
        blank=True,
        choices=[(1, "Very important"), (2, "Not as important")],
    )

    def __str__(self):  # __unicode__ on Python 2
        return self.headline

    class Meta:
        ordering = ("headline",)



from django.urls import path

from graphene.validation import DisableIntrospection

from ..views import GraphQLView
from .schema_view import schema


class View(GraphQLView):
    schema = schema


class NoIntrospectionView(View):
    validation_rules = (DisableIntrospection,)


class NoIntrospectionViewInherited(NoIntrospectionView):
    pass


urlpatterns = [
    path("graphql/", View.as_view()),
    path("graphql/validation/", View.as_view(validation_rules=(DisableIntrospection,))),
    path("graphql/validation/alternative/", NoIntrospectionView.as_view()),
    path("graphql/validation/inherited/", NoIntrospectionViewInherited.as_view()),
]



from functools import reduce

import pytest
from django.db import models
from django.db.models.query import QuerySet
from django_filters import FilterSet

import graphene
from graphene.relay import Node
from graphene_django import DjangoObjectType
from graphene_django.filter import ArrayFilter
from graphene_django.utils import DJANGO_FILTER_INSTALLED

from ...compat import ArrayField

pytestmark = []

if DJANGO_FILTER_INSTALLED:
    from graphene_django.filter import DjangoFilterConnectionField
else:
    pytestmark.append(
        pytest.mark.skipif(
            True, reason="django_filters not installed or not compatible"
        )
    )


class Event(models.Model):
    class Meta:
        ordering = ["pk"]

    name = models.CharField(max_length=50)
    tags = ArrayField(models.CharField(max_length=50))
    tag_ids = ArrayField(models.IntegerField())
    random_field = ArrayField(models.BooleanField())

    def __repr__(self):
        return f"Event [{self.name}]"


@pytest.fixture
def EventFilterSet():
    class EventFilterSet(FilterSet):
        class Meta:
            model = Event
            fields = {
                "name": ["exact", "contains"],
            }

        # Those are actually usable with our Query fixture below
        tags__contains = ArrayFilter(field_name="tags", lookup_expr="contains")
        tags__overlap = ArrayFilter(field_name="tags", lookup_expr="overlap")
        tags = ArrayFilter(field_name="tags", lookup_expr="exact")
        tags__len = ArrayFilter(
            field_name="tags", lookup_expr="len", input_type=graphene.Int
        )
        tags__len__in = ArrayFilter(
            field_name="tags",
            method="tags__len__in_filter",
            input_type=graphene.List(graphene.Int),
        )

        # Those are actually not usable and only to check type declarations
        tags_ids__contains = ArrayFilter(field_name="tag_ids", lookup_expr="contains")
        tags_ids__overlap = ArrayFilter(field_name="tag_ids", lookup_expr="overlap")
        tags_ids = ArrayFilter(field_name="tag_ids", lookup_expr="exact")
        random_field__contains = ArrayFilter(
            field_name="random_field", lookup_expr="contains"
        )
        random_field__overlap = ArrayFilter(
            field_name="random_field", lookup_expr="overlap"
        )
        random_field = ArrayFilter(field_name="random_field", lookup_expr="exact")

        def tags__len__in_filter(self, queryset, _name, value):
            if not value:
                return queryset.none()
            return reduce(
                lambda q1, q2: q1.union(q2),
                [queryset.filter(tags__len=v) for v in value],
            ).distinct()

    return EventFilterSet


@pytest.fixture
def EventType(EventFilterSet):
    class EventType(DjangoObjectType):
        class Meta:
            model = Event
            interfaces = (Node,)
            fields = "__all__"
            filterset_class = EventFilterSet

    return EventType


@pytest.fixture
def Query(EventType):
    """
    Note that we have to use a custom resolver to replicate the arrayfield filter behavior as
    we are running unit tests in sqlite which does not have ArrayFields.
    """

    events = [
        Event(name="Live Show", tags=["concert", "music", "rock"]),
        Event(name="Musical", tags=["movie", "music"]),
        Event(name="Ballet", tags=["concert", "dance"]),
        Event(name="Speech", tags=[]),
    ]

    class Query(graphene.ObjectType):
        events = DjangoFilterConnectionField(EventType)

        def resolve_events(self, info, **kwargs):
            class FakeQuerySet(QuerySet):
                def __init__(self, model=None):
                    self.model = Event
                    self.__store = list(events)

                def all(self):
                    return self

                def filter(self, **kwargs):
                    queryset = FakeQuerySet()
                    queryset.__store = list(self.__store)
                    if "tags__contains" in kwargs:
                        queryset.__store = list(
                            filter(
                                lambda e: set(kwargs["tags__contains"]).issubset(
                                    set(e.tags)
                                ),
                                queryset.__store,
                            )
                        )
                    if "tags__overlap" in kwargs:
                        queryset.__store = list(
                            filter(
                                lambda e: not set(kwargs["tags__overlap"]).isdisjoint(
                                    set(e.tags)
                                ),
                                queryset.__store,
                            )
                        )
                    if "tags__exact" in kwargs:
                        queryset.__store = list(
                            filter(
                                lambda e: set(kwargs["tags__exact"]) == set(e.tags),
                                queryset.__store,
                            )
                        )
                    if "tags__len" in kwargs:
                        queryset.__store = list(
                            filter(
                                lambda e: len(e.tags) == kwargs["tags__len"],
                                queryset.__store,
                            )
                        )
                    return queryset

                def union(self, *args):
                    queryset = FakeQuerySet()
                    queryset.__store = self.__store
                    for arg in args:
                        queryset.__store += arg.__store
                    return queryset

                def none(self):
                    queryset = FakeQuerySet()
                    queryset.__store = []
                    return queryset

                def count(self):
                    return len(self.__store)

                def distinct(self):
                    queryset = FakeQuerySet()
                    queryset.__store = []
                    for event in self.__store:
                        if event not in queryset.__store:
                            queryset.__store.append(event)
                    queryset.__store = sorted(queryset.__store, key=lambda e: e.name)
                    return queryset

                def __getitem__(self, index):
                    return self.__store[index]

            return FakeQuerySet()

    return Query


@pytest.fixture
def schema(Query):
    return graphene.Schema(query=Query)



import copy

from django.db import models
from pytest import raises
from rest_framework import serializers

import graphene

from ..serializer_converter import convert_serializer_field
from ..types import DictType


def _get_type(
    rest_framework_field, is_input=True, convert_choices_to_enum=True, **kwargs
):
    # prevents the following error:
    # AssertionError: The `source` argument is not meaningful when applied to a `child=` field.
    # Remove `source=` from the field declaration.
    # since we are reusing the same child in when testing the required attribute

    if "child" in kwargs:
        kwargs["child"] = copy.deepcopy(kwargs["child"])

    field = rest_framework_field(**kwargs)

    return convert_serializer_field(
        field, is_input=is_input, convert_choices_to_enum=convert_choices_to_enum
    )


def assert_conversion(rest_framework_field, graphene_field, **kwargs):
    graphene_type = _get_type(
        rest_framework_field, help_text="Custom Help Text", **kwargs
    )
    assert isinstance(graphene_type, graphene_field)

    graphene_type_required = _get_type(
        rest_framework_field, help_text="Custom Help Text", required=True, **kwargs
    )
    assert isinstance(graphene_type_required, graphene_field)

    return graphene_type


def test_should_unknown_rest_framework_field_raise_exception():
    with raises(Exception) as excinfo:
        convert_serializer_field(None)
    assert "Don't know how to convert the serializer field" in str(excinfo.value)


def test_should_char_convert_string():
    assert_conversion(serializers.CharField, graphene.String)


def test_should_email_convert_string():
    assert_conversion(serializers.EmailField, graphene.String)


def test_should_slug_convert_string():
    assert_conversion(serializers.SlugField, graphene.String)


def test_should_url_convert_string():
    assert_conversion(serializers.URLField, graphene.String)


def test_should_choice_convert_enum():
    field = assert_conversion(
        serializers.ChoiceField,
        graphene.Enum,
        choices=[("h", "Hello"), ("w", "World")],
        source="word",
    )
    assert field._meta.enum.__members__["H"].value == "h"
    assert field._meta.enum.__members__["H"].description == "Hello"
    assert field._meta.enum.__members__["W"].value == "w"
    assert field._meta.enum.__members__["W"].description == "World"


def test_should_choice_convert_string_if_enum_disabled():
    assert_conversion(
        serializers.ChoiceField,
        graphene.String,
        choices=[("h", "Hello"), ("w", "World")],
        source="word",
        convert_choices_to_enum=False,
    )


def test_should_base_field_convert_string():
    assert_conversion(serializers.Field, graphene.String)


def test_should_regex_convert_string():
    assert_conversion(serializers.RegexField, graphene.String, regex="[0-9]+")


def test_should_uuid_convert_string():
    if hasattr(serializers, "UUIDField"):
        assert_conversion(serializers.UUIDField, graphene.String)


def test_should_model_convert_field():
    class MyModelSerializer(serializers.ModelSerializer):
        class Meta:
            model = None
            fields = "__all__"

    assert_conversion(MyModelSerializer, graphene.Field, is_input=False)


def test_should_date_time_convert_datetime():
    assert_conversion(serializers.DateTimeField, graphene.types.datetime.DateTime)


def test_should_date_convert_date():
    assert_conversion(serializers.DateField, graphene.types.datetime.Date)


def test_should_time_convert_time():
    assert_conversion(serializers.TimeField, graphene.types.datetime.Time)


def test_should_integer_convert_int():
    assert_conversion(serializers.IntegerField, graphene.Int)


def test_should_boolean_convert_boolean():
    assert_conversion(serializers.BooleanField, graphene.Boolean)


def test_should_float_convert_float():
    assert_conversion(serializers.FloatField, graphene.Float)


def test_should_decimal_convert_decimal():
    assert_conversion(
        serializers.DecimalField, graphene.Decimal, max_digits=4, decimal_places=2
    )


def test_should_list_convert_to_list():
    class StringListField(serializers.ListField):
        child = serializers.CharField()

    field_a = assert_conversion(
        serializers.ListField,
        graphene.List,
        child=serializers.IntegerField(min_value=0, max_value=100),
    )

    assert field_a.of_type == graphene.Int

    field_b = assert_conversion(StringListField, graphene.List)

    assert field_b.of_type == graphene.String


def test_should_list_serializer_convert_to_list():
    class FooModel(models.Model):
        pass

    class ChildSerializer(serializers.ModelSerializer):
        class Meta:
            model = FooModel
            fields = "__all__"

    class ParentSerializer(serializers.ModelSerializer):
        child = ChildSerializer(many=True)

        class Meta:
            model = FooModel
            fields = "__all__"

    converted_type = convert_serializer_field(
        ParentSerializer().get_fields()["child"], is_input=True
    )
    assert isinstance(converted_type, graphene.List)

    converted_type = convert_serializer_field(
        ParentSerializer().get_fields()["child"], is_input=False
    )
    assert isinstance(converted_type, graphene.List)
    assert converted_type.of_type is None


def test_should_dict_convert_dict():
    assert_conversion(serializers.DictField, DictType)


def test_should_duration_convert_string():
    assert_conversion(serializers.DurationField, graphene.String)


def test_should_file_convert_string():
    assert_conversion(serializers.FileField, graphene.String)


def test_should_filepath_convert_string():
    assert_conversion(serializers.FilePathField, graphene.Enum, path="/")


def test_should_ip_convert_string():
    assert_conversion(serializers.IPAddressField, graphene.String)


def test_should_image_convert_string():
    assert_conversion(serializers.ImageField, graphene.String)


def test_should_json_convert_jsonstring():
    assert_conversion(serializers.JSONField, graphene.types.json.JSONString)


def test_should_multiplechoicefield_convert_to_list_of_enum():
    field = assert_conversion(
        serializers.MultipleChoiceField, graphene.List, choices=[1, 2, 3]
    )

    assert issubclass(field.of_type, graphene.Enum)
